#!/usr/bin/env python
# Scripts arguments:
# 1. Flow resolution
# 2. Scalar resolution
# 3. Dictionary for devices id: (mpi rank: device id)
# 4. Is the initial condition is perturbed
# 5. Is data output
import sys
msg = """
This example computes a two-phase flow.
Argments must be given in command line. For example:
mpirun -np 9 python RMI_hybrid.py "[129,129,257]" "[129,129,257]" "{0:0}" "True" "True"
"""
assert len(sys.argv) == 6, msg
USER_NB_ELEM_UW = eval(sys.argv[1])
USER_NB_ELEM_S = eval(sys.argv[2])
USER_RANK_DEVICE_ID = eval(sys.argv[3])
RANDOM_INIT = eval(sys.argv[4])
IS_OUTPUT = eval(sys.argv[5])
import hysop
# import hysop.gpu
# hysop.gpu.CL_PROFILE = True
from hysop.constants import np, HDF5, ASCII, HYSOP_MPI_REAL
from hysop.mpi.main_var import MPI, main_size, main_rank, main_comm
from hysop.tools.parameters import MPIParams, Discretization, IOParams
from hysop.problem.simulation import Simulation
from hysop.fields.variable_parameter import VariableParameter
from hysop.methods_keys import TimeIntegrator, Interpolation, Remesh,\
    Support, Splitting, MultiScale, MultiScale, SpaceDiscretisation, \
    Scales, dtCrit, ExtraArgs
from hysop.numerics.integrators.runge_kutta2 import RK2 as RK
from hysop.numerics.integrators.runge_kutta3 import RK3 as RK3
from hysop.numerics.finite_differences import FD_C_4
from hysop.numerics.remeshing import L6_4 as remesh_formula
from hysop.numerics.remeshing import Linear
from hysop.operator.advection import Advection
from hysop.operator.diffusion import Diffusion
from hysop.operator.stretching import Stretching
from hysop.operator.baroclinic import Baroclinic
from hysop.operator.penalization import PenalizeVorticity
from hysop.operator.poisson import Poisson
from hysop.operator.adapt_timestep import AdaptTimeStep
from hysop.operator.custom import CustomMonitor, CustomOp
from hysop.operator.redistribute_inter import RedistributeInter
from hysop.operator.redistribute_intra import RedistributeIntra
from hysop.gpu.gpu_transfer import DataTransfer
from hysop.domain.subsets import SubBox
from hysop.operator.hdf_io import HDF_Writer
from hysop.operator.energy_enstrophy import EnergyEnstrophy
import hysop.tools.numpywrappers as npw
from hysop.tools.profiler import Profiler, FProfiler
#from hysop.tools.io_utils import IO
#IO.set_default_path("/scratch_p/jmetancelin")
pi = np.pi
cos = np.cos
sin = np.sin
exp = np.exp
abs = np.abs
tanh = np.tanh

assert main_size > 1, "This example must run with at least 2 process"
TASK_UW = 1
TASK_SCALAR = 2
PROC_TASKS = [TASK_UW, ] * main_size
for p in USER_RANK_DEVICE_ID:
    PROC_TASKS[p] = TASK_SCALAR
try:
    DEVICE_ID = USER_RANK_DEVICE_ID[main_rank]
except KeyError:
    DEVICE_ID = None
out_freq = 10
# Physical parameters:
# Flow viscosity
VISCOSITY = 1e-4
AMPLITUDE = 0.02
PERIOD = pi
SIGMA = 0.1
NOISE = 1.


def computeVel(res, x, y, z, t):
    zz = exp(-z * z / SIGMA)
    res[0][...] = z * cos(PERIOD * x) * zz
    coef = -2. * AMPLITUDE / (PERIOD * SIGMA)
    res[0][...] *= coef
    res[1][...] = 0.
    res[2][...] = sin(PERIOD * x) * zz
    res[2][...] *= AMPLITUDE
    if RANDOM_INIT:
        from create_random_arrays import random_init
        randX, randY, randZ = random_init(res[0].shape, box.comm_task)
        res[0][...] *= (1. + zz * randX * NOISE)
        res[1][...] *= (1. + zz * randY * NOISE)
        res[2][...] *= (1. + zz * randZ * NOISE)
    return res


def computeVort(res, x, y, z, t):
    zz = exp(-z * z / SIGMA)
    coeff = 4.*AMPLITUDE*z*z/(PERIOD*SIGMA*SIGMA)
    res[0][...] = 0.
    res[1][...] = coeff * zz * cos(PERIOD * x)
    res[1][...] -= 2.*AMPLITUDE * zz * cos(PERIOD * x)/(PERIOD*SIGMA)
    res[1][...] -= AMPLITUDE*PERIOD * zz * cos(PERIOD * x)
    res[2][...] = 0.
    if RANDOM_INIT:
        from create_random_arrays import random_init
        randX, randY, randZ = random_init(res[0].shape, box.comm_task)
        res[0][...] *= (1. + zz * randX * NOISE)
        res[1][...] *= (1. + zz * randY * NOISE)
        res[2][...] *= (1. + zz * randZ * NOISE)
    return res

def initScal(res, x, y, z, t):
    res[0][...] = z
    return res

def phitorho(res, a):
    res[0][...] = np.tanh(a * 10.)
    res[0][...] += 2.
    return res

def initRho(res, x, y, z, t):
    return phitorho(res, z)


def func_scal_to_rho(simu, f_in, f_out):
    phitorho(f_out[0].data, f_in[0].data[0])


temp_maxvelo = npw.zeros((3, ))
maxvelo_values = npw.zeros((3, ))


def calc_maxvelo(simu, v):
    temp_maxvelo[0] = np.max(np.abs(v[0].data[0]))
    temp_maxvelo[1] = np.max(np.abs(v[0].data[1]))
    temp_maxvelo[2] = np.max(np.abs(v[0].data[2]))
    v[0].topology.comm.Allreduce(sendbuf=[temp_maxvelo, 3, HYSOP_MPI_REAL],
                                 recvbuf=[maxvelo_values, 3, HYSOP_MPI_REAL],
                                 op=MPI.MAX)
    return maxvelo_values

ctime = MPI.Wtime()
# Domain
box = hysop.Box(length=[2., 2., 4.], origin=[-1., -1., -2.],
                proc_tasks=PROC_TASKS)
bc_b = SubBox(length=[2., 2., 0.1], origin=[-1., -1., -2.], parent=box)
bc_t = SubBox(length=[2., 2., 0.1], origin=[-1., -1., 1.9], parent=box)
mpi_params = MPIParams(comm=box.comm_task, task_id=PROC_TASKS[main_rank])
mpi_params_S = MPIParams(comm=box.comm_task, task_id=TASK_SCALAR)
mpi_params_UW = MPIParams(comm=box.comm_task, task_id=TASK_UW)


# Fields
velo = hysop.Field(domain=box, formula=computeVel,
                   name='Velocity', is_vector=True)
vorti = hysop.Field(domain=box, formula=computeVort,
                    name='Vorticity', is_vector=True)
scal = hysop.Field(domain=box, formula=initScal,
                   name='Scalar', is_vector=False)
rho = hysop.Field(domain=box, formula=initRho,
                  name='Rho', is_vector=False)

data = {'dt': 0.001}
dt = VariableParameter(data)
simu = Simulation(tinit=0.0, tend=100., timeStep=0.001, iterMax=1000000)

# Flow discretizations:
d_uw = Discretization(USER_NB_ELEM_UW)
d_uw_ghosts = Discretization(USER_NB_ELEM_UW, [2, ] * 3)
# Scalar discretization
d_s = Discretization(USER_NB_ELEM_S)
# Velocity discretization for scalar advection
if USER_NB_ELEM_UW[0] == USER_NB_ELEM_S[0]:
    d_uw_for_s = Discretization(USER_NB_ELEM_UW)
else:
    d_uw_for_s = Discretization(USER_NB_ELEM_UW, [1, ] * 3)

# Topologies
topo_GPU_scal = None
topo_GPU_velo = None
topo_CPU_velo_fft = None
topo_CPU_velo_ghosts = None
topo_CPU_velo_scales = None
if box.is_on_task(TASK_UW):
    topo_CPU_velo_scales = box.create_topology(
        d_uw, dim=2, mpi_params=mpi_params_UW)
if box.is_on_task(TASK_SCALAR):
    topo_GPU_velo = box.create_topology(
        d_uw_for_s, dim=1, mpi_params=mpi_params_S)
    topo_GPU_scal = box.create_topology(
        d_s, dim=1, mpi_params=mpi_params_S)

# Operators
# GPU operators
advec_scal_method = {TimeIntegrator: RK,
                     Interpolation: Linear,
                     Remesh: remesh_formula,
                     Support: 'gpu_1k',
                     Splitting: 'o2',
                     MultiScale: Linear}
if PROC_TASKS.count(TASK_SCALAR) > 1:
    advec_scal_method[ExtraArgs] = {'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
else:
    advec_scal_method[ExtraArgs] = {'max_velocity': [1.2, 0.7, 0.7],
                                    'max_dt': 0.012,
                                    'device_id': DEVICE_ID,
                                    'device_type': 'gpu'}
advec_scal = Advection(velo,
                       discretization=topo_GPU_velo,
                       variables={scal: topo_GPU_scal},
                       mpi_params=mpi_params_S,
                       method=advec_scal_method)

# CPU operators
advec = Advection(velo,
                  discretization=topo_CPU_velo_scales,
                  variables={vorti: topo_CPU_velo_scales},
                  mpi_params=mpi_params_UW,
                  method={Scales: 'p_64', MultiScale: 'L4_4'})
stretch = Stretching(velo, vorti, discretization=d_uw_ghosts,
                     mpi_params=mpi_params_UW)
scal_to_rho = CustomOp([scal, ], [rho, ], func_scal_to_rho,
                       variables={scal: d_uw_ghosts,
                                  rho: d_uw_ghosts},
                       mpi_params=mpi_params_UW)
penal = PenalizeVorticity(velocity=velo, vorticity=vorti,
                          discretization=d_uw_ghosts,
                          obstacles=[bc_t, bc_b], coeff=1e8,
                          method={SpaceDiscretisation: FD_C_4})
baroclinic = Baroclinic(velo, vorti, rho, VISCOSITY,
                        discretization=d_uw_ghosts,
                        mpi_params=mpi_params_UW)
diffusion = Diffusion(variables={vorti: d_uw},
                      viscosity=VISCOSITY,
                      mpi_params=mpi_params_UW)
poisson = Poisson(velo, vorti, discretization=d_uw,
                  mpi_params=mpi_params_UW)
dt_output = IOParams(frequency=1, filename='dt.dat', fileformat=ASCII)
dt_adapt = AdaptTimeStep(velo, vorti,
                         simulation=simu,
                         time_range=[10, np.infty],
                         discretization=d_uw_ghosts,
                         method={TimeIntegrator: RK3,
                                 SpaceDiscretisation: FD_C_4,
                                 dtCrit: ['gradU']},
                         lcfl=0.1,
                         io_params=dt_output,
                         mpi_params=mpi_params_UW)

# Operators discretizations
if box.is_on_task(TASK_SCALAR):
    for op in (advec_scal, ):
        op.discretize()
if box.is_on_task(TASK_UW):
    for op in (advec, stretch, scal_to_rho, baroclinic, penal,
               diffusion, poisson, dt_adapt):
        op.discretize()

# Get some topologies
if box.is_on_task(TASK_UW):
    topo_CPU_velo_fft = poisson.discreteFields[velo].topology
    topo_CPU_velo_ghosts = stretch.discreteFields[velo].topology

if IS_OUTPUT:
    # Defining output operators
    p_velo = HDF_Writer(variables={velo: topo_CPU_velo_fft,
                                   vorti: topo_CPU_velo_fft},
                        mpi_params=mpi_params_UW,
                        io_params=IOParams(frequency=out_freq,
                                           filename='flow',
                                           fileformat=HDF5))
    p_rho = HDF_Writer(variables={rho: topo_CPU_velo_ghosts},
                       mpi_params=mpi_params_UW,
                       io_params=IOParams(frequency=out_freq,
                                          filename='rho',
                                          fileformat=HDF5))
    p_scal = HDF_Writer(variables={scal: topo_GPU_scal},
                        var_names={scal: 'Scalar'},
                        mpi_params=mpi_params_S,
                        io_params=IOParams(frequency=out_freq,
                                           filename='scal',
                                           fileformat=HDF5))
    p_scal.name += '_(Scalar)'
    energy = EnergyEnstrophy(velocity=velo,
                             vorticity=vorti,
                             discretization=topo_CPU_velo_fft,
                             mpi_params=mpi_params_UW,
                             io_params=IOParams(frequency=1,
                                                filename='energy.dat',
                                                fileformat=ASCII))
    maxvelo = CustomMonitor(variables={velo: topo_CPU_velo_scales},
                            function=calc_maxvelo,
                            res_shape=3,
                            mpi_params=mpi_params_UW,
                            io_params=IOParams(frequency=1,
                                               filename='maxvelo.dat',
                                               fileformat=ASCII))
    if box.is_on_task(TASK_UW):
        for op in (p_velo, p_rho, energy, maxvelo):
            op.discretize()
    if box.is_on_task(TASK_SCALAR):
        for op in (p_scal, ):
            op.discretize()

# Redistribute operators
# CPU redistributes
advecToGhost_vorti = RedistributeIntra(variables=[vorti],
                                       source=advec, target=stretch,
                                       mpi_params=mpi_params_UW)
advecToGhost_vorti.name += '_advec2Ghost_W'
fftToGhost_velo = RedistributeIntra(variables=[velo],
                                    source=poisson, target=stretch,
                                    run_till=[stretch, dt_adapt],
                                    mpi_params=mpi_params_UW)
fftToGhost_velo.name += '_advec2Ghost_U'
fftToGhost_vorti = RedistributeIntra(variables=[vorti],
                                     source=poisson, target=stretch,
                                     run_till=[stretch, dt_adapt],
                                     mpi_params=mpi_params_UW)
fftToGhost_vorti.name += '_fft2Ghost_W'
fftToScales_velo = RedistributeIntra(variables=[velo],
                                     source=poisson, target=advec,
                                     mpi_params=mpi_params_UW)
fftToScales_velo.name += '_fft2Scales_U'
fftToScales_vorti = RedistributeIntra(variables=[vorti],
                                      source=poisson, target=advec,
                                      mpi_params=mpi_params_UW)
fftToScales_vorti.name += '_fft2Scales_W'
ghostToFFT_vorti = RedistributeIntra(variables=[vorti],
                                     source=baroclinic, target=diffusion,
                                     mpi_params=mpi_params_UW)
ghostToFFT_vorti.name += '_ghost2FFT_W'
ghostToFFT_velo = RedistributeIntra(variables=[velo],
                                    source=penal, target=poisson,
                                    mpi_params=mpi_params_UW)
ghostToFFT_velo.name += '_ghost2FFT_U'
toDev = DataTransfer(source=topo_GPU_velo, target=advec_scal,
                     variables={velo: topo_GPU_velo},
                     mpi_params=mpi_params_S)
toDev.name += '_ToDev_U'

redistrib = RedistributeInter(variables=[velo],
                              parent=main_comm,
                              source=topo_CPU_velo_fft, target=topo_GPU_velo,
                              source_id=TASK_UW, target_id=TASK_SCALAR,
                              run_till=[toDev])
redistrib.name += '_CPU2GPU_U'
redistrib_scal = RedistributeInter(variables=[scal],
                                   parent=main_comm,
                                   source=topo_GPU_velo,
                                   target=topo_CPU_velo_ghosts,
                                   source_id=TASK_SCALAR, target_id=TASK_UW,
                                   run_till=[scal_to_rho])
redistrib_scal.name += '_GPU2CPU_Scal'
toHost = DataTransfer(source=topo_GPU_scal, target=p_scal,
                      variables={scal: topo_GPU_scal},
                      mpi_params=mpi_params_S,
                      run_till=[redistrib_scal, ])
toHost.name += '_ToHost_S'

# Operators setup
if box.is_on_task(TASK_SCALAR):
    for op in (advec_scal, ):
        op.setup()
    if IS_OUTPUT:
        for op in (p_scal, toHost):
            op.setup()
    for op in (toDev, ):
        op.setup()
if box.is_on_task(TASK_UW):
    for op in (advec, stretch, scal_to_rho, baroclinic, penal,
               diffusion, poisson, dt_adapt):
        op.setup()
    if IS_OUTPUT:
        for op in (p_velo, p_rho, energy, maxvelo):
            op.setup()
    for op in (advecToGhost_vorti, fftToGhost_velo,
               fftToGhost_vorti, fftToScales_velo,
               fftToScales_vorti, ghostToFFT_vorti,
               ghostToFFT_velo):
        op.setup()
# Wait for all operators setup before setup the intra-comm redistribute
main_comm.Barrier()
redistrib.setup()
redistrib_scal.setup()

# Operators list
if IS_OUTPUT:
    operators_list = [redistrib,
                      toDev, advec_scal,
                      toHost, p_scal,
                      advec, advecToGhost_vorti, stretch, baroclinic,
                      redistrib_scal,
                      scal_to_rho, penal,
                      ghostToFFT_velo, ghostToFFT_vorti,
                      diffusion, poisson,
                      p_velo, p_rho, energy, maxvelo,
                      fftToGhost_velo, fftToScales_velo,
                      fftToScales_vorti, dt_adapt]
else:
    operators_list = [redistrib,
                      toDev, advec_scal,
                      toHost,
                      advec, advecToGhost_vorti, stretch, baroclinic,
                      redistrib_scal,
                      scal_to_rho, penal,
                      ghostToFFT_velo, ghostToFFT_vorti,
                      diffusion, poisson,
                      fftToGhost_velo, fftToScales_velo,
                      fftToScales_vorti, dt_adapt]


# Fields initializations
if box.is_on_task(TASK_SCALAR):
    scal.initialize(topo=topo_GPU_scal)
    advec_dirX = advec_scal.advec_dir[0].discrete_op
    gpu_scal = advec_dirX.fields_on_grid[0]
    gpu_pscal = advec_dirX.fields_on_part[gpu_scal][0]
    if IS_OUTPUT:
        p_scal.apply(simu)
if box.is_on_task(TASK_UW):
    velo.initialize(topo=topo_CPU_velo_ghosts)
    vorti.initialize(topo=topo_CPU_velo_ghosts)
    rho.initialize(topo=topo_CPU_velo_ghosts)
    bc_b.discretize(topo_CPU_velo_ghosts)
    bc_t.discretize(topo_CPU_velo_ghosts)
    penal.apply(simu)
    ghostToFFT_vorti.apply(simu)
    ghostToFFT_vorti.wait()
    ghostToFFT_velo.apply(simu)
    ghostToFFT_velo.wait()
    fftToScales_vorti.apply(simu)
    fftToScales_vorti.wait()
    fftToScales_velo.apply(simu)
    fftToScales_velo.wait()
    baroclinic.initialize_velocity()
    if IS_OUTPUT:
        p_velo.apply(simu)

simu.initialize()
setup_time = MPI.Wtime() - ctime
main_comm.Barrier()

# Solve
solve_time = FProfiler("Solve")
while not simu.isOver:
    ctime = MPI.Wtime()
    if main_rank == 0:
        simu.printState()
    for op in operators_list:
        if box.is_on_task(op.task_id()):
            op.apply(simu)
    if box.is_on_task(TASK_SCALAR):
        # Wait gpu operations on scalar
        advec_scal.advec_dir[0].discreteFields[scal].wait()
    solve_time += MPI.Wtime() - ctime
    dt_adapt.wait()
    # Synchronize threads
    main_comm.Barrier()
    simu.advance()

main_comm.Barrier()
nb_ite = simu.currentIteration
simu.finalize()


if IS_OUTPUT:
    if box.is_on_task(TASK_SCALAR) and \
       simu.time != p_scal._last_written_time:
        p_scal.apply(simu)
    if box.is_on_task(TASK_UW) and \
       simu.time != p_velo._last_written_time:
        p_velo.apply(simu)

prof = Profiler(None, box.comm_task)
prof += solve_time
for op in operators_list:
    if box.is_on_task(op.task_id()):
        op.finalize()
        prof += op.profiler
for v in (velo, vorti, rho, scal):
    prof += v.profiler
prof.summarize()

for i in xrange(main_size):
    if i == main_rank:
        print "Solving Time:", solve_time
        print prof
    main_comm.Barrier()

velo.finalize()
if box.is_on_task(TASK_SCALAR):
    scal.finalize()
if box.is_on_task(TASK_UW):
    vorti.finalize()
main_comm.Barrier()
